#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributers.

from fenrirscreenreader.core.i18n import _

import re
import time
import threading


class command():
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        # Use commandBuffer like other commands
        if 'progressMonitoring' not in self.env['commandBuffer']:
            # Check if progress monitoring should be enabled by default from
            # settings
            try:
                defaultEnabled = self.env['runtime']['settingsManager'].getSettingAsBool(
                    'sound', 'progressMonitoring')
            except Exception as e:
                # If setting doesn't exist, default to False
                defaultEnabled = False
            self.env['commandBuffer']['progressMonitoring'] = defaultEnabled
            self.env['commandBuffer']['lastProgressTime'] = 0
            self.env['commandBuffer']['lastProgressValue'] = -1

    def shutdown(self):
        self.stopProgressMonitoring()

    def getDescription(self):
        return _('Toggle progress bar monitoring with ascending tones')

    def run(self):
        # Check if commandBuffer exists
        if 'progressMonitoring' not in self.env['commandBuffer']:
            self.env['commandBuffer']['progressMonitoring'] = False
            self.env['commandBuffer']['lastProgressTime'] = 0
            self.env['commandBuffer']['lastProgressValue'] = -1

        if self.env['commandBuffer']['progressMonitoring']:
            self.stopProgressMonitoring()
            self.env['runtime']['outputManager'].presentText(
                _("Progress monitoring disabled"), interrupt=True)
        else:
            self.startProgressMonitoring()
            self.env['runtime']['outputManager'].presentText(
                _("Progress monitoring enabled"), interrupt=True)

    def startProgressMonitoring(self):
        self.env['commandBuffer']['progressMonitoring'] = True
        self.env['commandBuffer']['lastProgressTime'] = time.time()
        self.env['commandBuffer']['lastProgressValue'] = -1
        # Don't control speech - let user decide with silence_until_prompt

    def stopProgressMonitoring(self):
        self.env['commandBuffer']['progressMonitoring'] = False
        # Don't control speech - progress monitor is beep-only

    def detectProgress(self, text):
        if not self.env['commandBuffer']['progressMonitoring']:
            return

        # Skip progress detection if current screen looks like a prompt
        if self.isCurrentLinePrompt():
            return

        currentTime = time.time()

        # Pattern 1: Percentage (50%, 25.5%, etc.)
        percentMatch = re.search(r'(\d+(?:\.\d+)?)\s*%', text)
        if percentMatch:
            percentage = float(percentMatch.group(1))
            if percentage != self.env['commandBuffer']['lastProgressValue']:
                self.playProgressTone(percentage)
                self.env['commandBuffer']['lastProgressValue'] = percentage
                self.env['commandBuffer']['lastProgressTime'] = currentTime
            return

        # Pattern 2: Fraction (15/100, 3 of 10, etc.)
        fractionMatch = re.search(r'(\d+)\s*(?:of|/)\s*(\d+)', text)
        if fractionMatch:
            current = int(fractionMatch.group(1))
            total = int(fractionMatch.group(2))
            if total > 0:
                percentage = (current / total) * 100
                if percentage != self.env['commandBuffer']['lastProgressValue']:
                    self.playProgressTone(percentage)
                    self.env['commandBuffer']['lastProgressValue'] = percentage
                    self.env['commandBuffer']['lastProgressTime'] = currentTime
                return

        # Pattern 3: Progress bars ([####    ], [====>   ], etc.)
        # Improved pattern to avoid matching IRC channels like [#channel]
        barMatch = re.search(r'\[([#=\-\*]+)([\s\.]*)\]', text)
        if barMatch:
            filled = len(barMatch.group(1))
            unfilled = len(barMatch.group(2))
            total = filled + unfilled
            # Require at least 2 progress chars total and unfilled portion must
            # be spaces/dots
            if total >= 2 and (
                not barMatch.group(2) or re.match(
                    r'^[\s\.]*$',
                    barMatch.group(2))):
                percentage = (filled / total) * 100
                if percentage != self.env['commandBuffer']['lastProgressValue']:
                    self.playProgressTone(percentage)
                    self.env['commandBuffer']['lastProgressValue'] = percentage
                    self.env['commandBuffer']['lastProgressTime'] = currentTime
                return

        # Pattern 4: Generic activity indicators (Loading..., Working..., etc.)
        activityPattern = re.search(
            r'(loading|processing|working|installing|downloading|compiling|building).*\.{2,}',
            text,
            re.IGNORECASE)
        if activityPattern:
            # Play a steady beep every 2 seconds for ongoing activity
            if currentTime - \
                    self.env['commandBuffer']['lastProgressTime'] >= 2.0:
                self.playActivityBeep()
                self.env['commandBuffer']['lastProgressTime'] = currentTime

    def playProgressTone(self, percentage):
        # Map 0-100% to 400-1200Hz frequency range
        frequency = 400 + (percentage * 8)
        frequency = max(400, min(1200, frequency))  # Clamp to safe range
        self.env['runtime']['outputManager'].playFrequence(
            frequency, 0.15, interrupt=False)

    def playActivityBeep(self):
        # Single tone for generic activity
        self.env['runtime']['outputManager'].playFrequence(
            800, 0.1, interrupt=False)

    def isCurrentLinePrompt(self):
        """Check if the current line looks like a standalone prompt (not command with progress)"""
        import re

        try:
            # Get the current screen content
            if not self.env['screen']['newContentText']:
                return False

            lines = self.env['screen']['newContentText'].split('\n')
            if not lines:
                return False

            # Check the last line (most common) and current cursor line for
            # prompt patterns
            linesToCheck = []

            # Add last line (most common for prompts)
            if lines:
                linesToCheck.append(lines[-1])

            # Add current cursor line if different from last line
            if (self.env['screen']['newCursor']['y'] < len(lines) and
                    self.env['screen']['newCursor']['y'] != len(lines) - 1):
                linesToCheck.append(
                    lines[self.env['screen']['newCursor']['y']])

            # Standalone prompt patterns (no commands mixed in)
            standalonePromptPatterns = [
                r'^\s*\$\s*$',  # Just $ (with whitespace)
                r'^\s*#\s*$',   # Just # (with whitespace)
                r'^\s*>\s*$',   # Just > (with whitespace)
                r'^\[.*\]\s*[\\\$#>]\s*$',  # [path]$ without commands
                r'^[a-zA-Z0-9._-]+[\\\$#>]\s*$',  # bash-5.1$ without commands

                # Interactive prompt patterns (these ARE standalone)
                r'.*\?\s*\[[YyNn]/[YyNn]\]\s*$',  # ? [Y/n] or ? [y/N] style
                r'.*\?\s*\[[Yy]es/[Nn]o\]\s*$',  # ? [Yes/No] style
                # "continue? [Y/n]" style
                r'.*continue\?\s*\[[YyNn]/[YyNn]\].*$',
                r'^::.*\?\s*\[[YyNn]/[YyNn]\].*$',  # pacman style prompts

                # Authentication prompts (these ARE standalone)
                r'^\[[Ss]udo\]\s*[Pp]assword\s*for\s+.*:\s*$',  # [sudo] password
                r'^[Pp]assword\s*:\s*$',  # Password:
                r'.*[Pp]assword\s*:\s*$',  # general password prompts

                # Continuation prompts (these ARE standalone)
                r'^[Pp]ress\s+any\s+key\s+to\s+continue.*$',  # Press any key
                r'^[Aa]re\s+you\s+sure\?\s*.*$',  # Are you sure?
            ]

            for line in linesToCheck:
                line = line.strip()
                if not line:
                    continue

                # Check if this line contains both a prompt AND other content (like commands)
                # If so, don't treat it as a standalone prompt
                hasPromptMarker = bool(
                    re.search(
                        r'.*@.*[\\\$#>]',
                        line) or re.search(
                        r'^\[.*\]\s*[\\\$#>]',
                        line))
                if hasPromptMarker:
                    # If line has prompt marker but also has significant content after it,
                    # it's a command line, not a standalone prompt
                    promptEnd = max(
                        line.rfind('$'),
                        line.rfind('#'),
                        line.rfind('>'),
                        line.rfind('\\')
                    )
                    if promptEnd >= 0 and promptEnd < len(
                            line) - 5:  # More than just whitespace after prompt
                        continue  # This is a command line, not a standalone prompt

                for pattern in standalonePromptPatterns:
                    try:
                        if re.search(pattern, line):
                            return True
                    except re.error:
                        continue

            return False

        except Exception as e:
            # If anything fails, assume it's not a prompt to be safe
            return False

    def setCallback(self, callback):
        pass
