#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _

import subprocess
import time


class command():
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        self.testMessage = "This is a voice test. The quick brown fox jumps over the lazy dog."
        self.modules = []
        self.voices = []
        self.moduleIndex = 0
        self.voiceIndex = 0
        self.browserActive = False
        self.originalBindings = {}
        self.lastAnnounceTime = 0

    def shutdown(self):
        pass

    def getDescription(self):
        return "Interactive voice browser with arrow key navigation"

    def run(self):
        if self.browserActive:
            self.exitVoiceBrowser()
            return

        self.env['runtime']['outputManager'].presentText(
            "Starting voice browser", interrupt=True)

        # Load modules
        self.modules = self.getSpeechdModules()
        if not self.modules:
            self.env['runtime']['outputManager'].presentText(
                "No speech modules found", interrupt=True)
            self.env['runtime']['outputManager'].playSound('Error')
            return

        # Set current module
        currentModule = self.env['runtime']['settingsManager'].getSetting(
            'speech', 'module')
        if currentModule and currentModule in self.modules:
            self.moduleIndex = self.modules.index(currentModule)

        # Load voices
        self.loadVoicesForCurrentModule()

        # Set current voice
        currentVoice = self.env['runtime']['settingsManager'].getSetting(
            'speech', 'voice')
        if currentVoice and currentVoice in self.voices:
            self.voiceIndex = self.voices.index(currentVoice)

        # Enter browser mode
        self.enterVoiceBrowser()
        self.announceCurrentSelection()

    def enterVoiceBrowser(self):
        """Enter voice browser mode with custom key bindings"""
        self.browserActive = True

        # Store original bindings
        self.originalBindings = self.env['bindings'].copy()

        # Override navigation keys for voice browsing
        # Use lambda to capture self and create bound methods
        self.env['bindings'][str([1, ['KEY_UP']])] = 'VOICE_BROWSER_PREV_VOICE'
        self.env['bindings'][str([1, ['KEY_DOWN']])
                             ] = 'VOICE_BROWSER_NEXT_VOICE'
        self.env['bindings'][str([1, ['KEY_LEFT']])
                             ] = 'VOICE_BROWSER_PREV_MODULE'
        self.env['bindings'][str([1, ['KEY_RIGHT']])
                             ] = 'VOICE_BROWSER_NEXT_MODULE'
        self.env['bindings'][str([1, ['KEY_ENTER']])] = 'VOICE_BROWSER_TEST'
        self.env['bindings'][str([1, ['KEY_SPACE']])] = 'VOICE_BROWSER_APPLY'
        self.env['bindings'][str([1, ['KEY_ESC']])] = 'VOICE_BROWSER_EXIT'

        # Register browser commands
        self.registerBrowserCommands()

        self.env['runtime']['outputManager'].presentText(
            "Voice browser active", interrupt=True)
        self.env['runtime']['outputManager'].presentText(
            "Up/Down=voices, Left/Right=modules, Enter=test, Space=apply, Esc=exit",
            interrupt=True)

    def registerBrowserCommands(self):
        """Register voice browser commands with the command manager"""
        # Create command objects for voice browser actions
        commandManager = self.env['runtime']['commandManager']

        # This is a hack - we'll store references to our methods in the environment
        # so the key handlers can call them
        if 'voiceBrowserInstance' not in self.env['runtime']:
            self.env['runtime']['voiceBrowserInstance'] = self

    def exitVoiceBrowser(self):
        """Exit voice browser and restore normal key bindings"""
        if not self.browserActive:
            return

        self.browserActive = False

        # Restore original bindings
        self.env['bindings'] = self.originalBindings

        # Clean up
        if 'voiceBrowserInstance' in self.env['runtime']:
            del self.env['runtime']['voiceBrowserInstance']

        self.env['runtime']['outputManager'].presentText(
            "Voice browser exited", interrupt=True)

    def loadVoicesForCurrentModule(self):
        """Load voices for current module"""
        if self.moduleIndex < len(self.modules):
            module = self.modules[self.moduleIndex]
            self.voices = self.getModuleVoices(module)
            self.voiceIndex = 0  # Reset to first voice when changing modules

    def announceCurrentSelection(self):
        """Announce current module and voice"""
        # Throttle announcements to avoid spam
        now = time.time()
        if now - self.lastAnnounceTime < 0.3:
            return
        self.lastAnnounceTime = now

        if not self.modules or self.moduleIndex >= len(self.modules):
            return

        module = self.modules[self.moduleIndex]
        if self.voices and self.voiceIndex < len(self.voices):
            voice = self.voices[self.voiceIndex]
            self.env['runtime']['outputManager'].presentText(
                f"{module}: {voice} ({self.voiceIndex + 1}/{len(self.voices)})", interrupt=True
            )
        else:
            self.env['runtime']['outputManager'].presentText(
                f"{module}: No voices", interrupt=True)

    def nextVoice(self):
        """Move to next voice"""
        if not self.voices:
            return
        self.voiceIndex = (self.voiceIndex + 1) % len(self.voices)
        self.announceCurrentSelection()

    def prevVoice(self):
        """Move to previous voice"""
        if not self.voices:
            return
        self.voiceIndex = (self.voiceIndex - 1) % len(self.voices)
        self.announceCurrentSelection()

    def nextModule(self):
        """Move to next module"""
        self.moduleIndex = (self.moduleIndex + 1) % len(self.modules)
        self.loadVoicesForCurrentModule()
        self.announceCurrentSelection()

    def prevModule(self):
        """Move to previous module"""
        self.moduleIndex = (self.moduleIndex - 1) % len(self.modules)
        self.loadVoicesForCurrentModule()
        self.announceCurrentSelection()

    def testVoice(self):
        """Test current voice"""
        if not self.voices or self.voiceIndex >= len(self.voices):
            self.env['runtime']['outputManager'].presentText(
                "No voice selected", interrupt=True)
            return

        module = self.modules[self.moduleIndex]
        voice = self.voices[self.voiceIndex]

        self.env['runtime']['outputManager'].presentText(
            "Testing...", interrupt=True)
        if self.previewVoice(module, voice):
            # Store for apply command
            self.env['commandBuffer']['lastTestedModule'] = module
            self.env['commandBuffer']['lastTestedVoice'] = voice
            self.env['runtime']['outputManager'].playSound('Accept')
        else:
            self.env['runtime']['outputManager'].playSound('Error')

    def applyVoice(self):
        """Apply current voice to Fenrir"""
        if not self.voices or self.voiceIndex >= len(self.voices):
            return

        module = self.modules[self.moduleIndex]
        voice = self.voices[self.voiceIndex]

        try:
            settingsManager = self.env['runtime']['settingsManager']
            settingsManager.settings['speech']['driver'] = 'speechdDriver'
            settingsManager.settings['speech']['module'] = module
            settingsManager.settings['speech']['voice'] = voice

            if 'speechDriver' in self.env['runtime']:
                speechDriver = self.env['runtime']['speechDriver']
                speechDriver.shutdown()
                speechDriver.initialize(self.env)

                self.env['runtime']['outputManager'].presentText(
                    "Voice applied!", interrupt=True)
                self.env['runtime']['outputManager'].playSound('Accept')

        except Exception as e:
            self.env['runtime']['outputManager'].presentText(
                "Apply failed", interrupt=True)
            self.env['runtime']['outputManager'].playSound('Error')

    def previewVoice(self, module, voice):
        """Test voice with spd-say"""
        try:
            cmd = ['spd-say', '-o', module, '-y', voice, self.testMessage]
            result = subprocess.run(cmd, timeout=10)
            return result.returncode == 0
        except Exception:
            return False

    def getSpeechdModules(self):
        """Get available speech modules"""
        try:
            result = subprocess.run(
                ['spd-say', '-O'], capture_output=True, text=True, timeout=10)
            if result.returncode == 0:
                lines = result.stdout.strip().split('\n')
                return [line.strip() for line in lines[1:] if line.strip()]
        except Exception:
            pass
        return []

    def getModuleVoices(self, module):
        """Get voices for module"""
        try:
            result = subprocess.run(
                ['spd-say', '-o', module, '-L'], capture_output=True, text=True, timeout=10)
            if result.returncode == 0:
                lines = result.stdout.strip().split('\n')
                voices = []
                for line in lines[1:]:
                    if not line.strip():
                        continue
                    if module.lower() == 'espeak-ng':
                        voice = self.processEspeakVoice(line)
                        if voice:
                            voices.append(voice)
                    else:
                        voices.append(line.strip())
                return voices
        except Exception:
            pass
        return []

    def processEspeakVoice(self, voiceLine):
        """Process espeak voice format"""
        parts = [p for p in voiceLine.split() if p]
        if len(parts) < 2:
            return None
        langCode = parts[-2].lower()
        variant = parts[-1].lower()
        return f"{langCode}+{variant}" if variant and variant != 'none' else langCode

    def setCallback(self, callback):
        pass
