#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _

import subprocess
import threading
import time


class command():
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        self.testMessage = "Voice test: The quick brown fox jumps over the lazy dog."

    def shutdown(self):
        pass

    def getDescription(self):
        return "Safe voice browser - cycles through voices without hanging"

    def run(self):
        try:
            self.env['runtime']['outputManager'].presentText(
                "Starting safe voice browser", interrupt=True)

            # Get modules with timeout protection
            modules = self.getSpeechdModulesWithTimeout()
            if not modules:
                self.env['runtime']['outputManager'].presentText(
                    "No speech modules found", interrupt=True)
                return

            # Get current position from commandBuffer or start fresh
            moduleIndex = self.env['commandBuffer'].get(
                'safeBrowserModuleIndex', 0)
            voiceIndex = self.env['commandBuffer'].get(
                'safeBrowserVoiceIndex', 0)

            # Ensure valid module index
            if moduleIndex >= len(modules):
                moduleIndex = 0

            currentModule = modules[moduleIndex]
            self.env['runtime']['outputManager'].presentText(
                f"Loading voices for {currentModule}...", interrupt=True)

            # Get voices with timeout protection
            voices = self.getModuleVoicesWithTimeout(currentModule)
            if not voices:
                self.env['runtime']['outputManager'].presentText(
                    f"No voices in {currentModule}, trying next module", interrupt=True)
                moduleIndex = (moduleIndex + 1) % len(modules)
                self.env['commandBuffer']['safeBrowserModuleIndex'] = moduleIndex
                self.env['commandBuffer']['safeBrowserVoiceIndex'] = 0
                return

            # Ensure valid voice index
            if voiceIndex >= len(voices):
                voiceIndex = 0

            currentVoice = voices[voiceIndex]

            # Announce current selection
            self.env['runtime']['outputManager'].presentText(
                f"Module: {currentModule} ({moduleIndex + 1}/{len(modules)})", interrupt=True
            )
            self.env['runtime']['outputManager'].presentText(
                f"Voice: {currentVoice} ({voiceIndex + 1}/{len(voices)})", interrupt=True
            )

            # Test voice in background thread to avoid blocking
            self.env['runtime']['outputManager'].presentText(
                "Testing voice...", interrupt=True)

            # Use threading to prevent freezing
            testThread = threading.Thread(
                target=self.testVoiceAsync, args=(
                    currentModule, currentVoice))
            testThread.daemon = True
            testThread.start()

            # Store tested voice for apply command
            self.env['commandBuffer']['lastTestedModule'] = currentModule
            self.env['commandBuffer']['lastTestedVoice'] = currentVoice

            # Advance to next voice for next run
            voiceIndex += 1
            if voiceIndex >= len(voices):
                voiceIndex = 0
                moduleIndex = (moduleIndex + 1) % len(modules)

            # Store position for next run
            self.env['commandBuffer']['safeBrowserModuleIndex'] = moduleIndex
            self.env['commandBuffer']['safeBrowserVoiceIndex'] = voiceIndex

            # Give instructions
            self.env['runtime']['outputManager'].presentText(
                "Run again for next voice, or use apply voice command", interrupt=True)

        except Exception as e:
            self.env['runtime']['outputManager'].presentText(
                f"Voice browser error: {str(e)}", interrupt=True)
            self.env['runtime']['outputManager'].playSound('Error')

    def testVoiceAsync(self, module, voice):
        """Test voice in background thread to avoid blocking"""
        try:
            # Run with strict timeout
            cmd = ['spd-say', '-o', module, '-y', voice, self.testMessage]
            result = subprocess.run(cmd, timeout=5, capture_output=True)

            # Schedule success sound for main thread
            if result.returncode == 0:
                # We can't call outputManager from background thread safely
                # So we'll just let the main thread handle feedback
                pass

        except subprocess.TimeoutExpired:
            # Voice test timed out - this is okay, don't crash
            pass
        except Exception:
            # Any other error - also okay, don't crash
            pass

    def getSpeechdModulesWithTimeout(self):
        """Get speech modules with timeout protection"""
        try:
            result = subprocess.run(
                ['spd-say', '-O'], capture_output=True, text=True, timeout=3)
            if result.returncode == 0:
                lines = result.stdout.strip().split('\n')
                modules = [line.strip() for line in lines[1:] if line.strip()]
                # Limit to first 10 modules to prevent overload
                return modules[:10]
        except subprocess.TimeoutExpired:
            self.env['runtime']['outputManager'].presentText(
                "Module detection timed out", interrupt=True)
        except Exception as e:
            self.env['runtime']['outputManager'].presentText(
                f"Module detection failed: {str(e)}", interrupt=True)
        return []

    def getModuleVoicesWithTimeout(self, module):
        """Get voices with timeout and limits"""
        try:
            result = subprocess.run(
                ['spd-say', '-o', module, '-L'], capture_output=True, text=True, timeout=5)
            if result.returncode == 0:
                lines = result.stdout.strip().split('\n')
                voices = []
                for line in lines[1:]:
                    if not line.strip():
                        continue
                    if module.lower() == 'espeak-ng':
                        voice = self.processEspeakVoice(line)
                        if voice:
                            voices.append(voice)
                    else:
                        voices.append(line.strip())

                # Limit voice count to prevent memory issues
                if len(voices) > 1000:
                    self.env['runtime']['outputManager'].presentText(
                        f"Found {len(voices)} voices, limiting to first 1000", interrupt=True)
                    voices = voices[:1000]

                return voices
        except subprocess.TimeoutExpired:
            self.env['runtime']['outputManager'].presentText(
                f"Voice detection for {module} timed out", interrupt=True)
        except Exception as e:
            self.env['runtime']['outputManager'].presentText(
                f"Voice detection failed: {str(e)}", interrupt=True)
        return []

    def processEspeakVoice(self, voiceLine):
        """Process espeak voice format"""
        try:
            parts = [p for p in voiceLine.split() if p]
            if len(parts) < 2:
                return None
            langCode = parts[-2].lower()
            variant = parts[-1].lower()
            return f"{langCode}+{variant}" if variant and variant != 'none' else langCode
        except Exception:
            return None

    def setCallback(self, callback):
        pass
