#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _

import os
import importlib.util

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), '..', 'config_base.py')
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
ConfigCommand = _module.ConfigCommand


class command(ConfigCommand):
    def __init__(self):
        super().__init__()

    def getDescription(self):
        return "Select keyboard layout (desktop or laptop)"

    def run(self):
        currentLayout = self.getSetting(
            'keyboard', 'keyboardLayout', 'desktop')

        # Present current layout
        self.presentText(f"Current keyboard layout: {currentLayout}")

        # Find available keyboard layouts
        keyboardPath = '/etc/fenrirscreenreader/keyboard'
        if not os.path.isdir(keyboardPath):
            # Development path
            keyboardPath = os.path.join(
                os.path.dirname(
                    self.settingsFile),
                '..',
                'keyboard')

        availableLayouts = self.getAvailableLayouts(keyboardPath)

        if len(availableLayouts) > 1:
            # Cycle through available layouts
            try:
                currentIndex = availableLayouts.index(currentLayout)
                nextIndex = (currentIndex + 1) % len(availableLayouts)
                newLayout = availableLayouts[nextIndex]
            except ValueError:
                # Current layout not found, use first available
                newLayout = availableLayouts[0]

            success = self.setSetting('keyboard', 'keyboardLayout', newLayout)

            if success:
                self.presentText(f"Keyboard layout changed to: {newLayout}")
                self.presentText(
                    "Please restart Fenrir for this change to take effect.")
                self.playSound('Accept')
            else:
                self.presentText("Failed to change keyboard layout")
                self.playSound('Error')
        else:
            self.presentText("Only default keyboard layout is available")
            self.playSound('Cancel')

    def getAvailableLayouts(self, keyboardPath):
        """Find available keyboard layout files"""
        layouts = []

        if os.path.isdir(keyboardPath):
            try:
                for file in os.listdir(keyboardPath):
                    if file.endswith('.conf') and not file.startswith('.'):
                        layoutName = file[:-5]  # Remove .conf extension
                        layouts.append(layoutName)
            except Exception:
                pass

        # Ensure we have at least the default layouts
        if not layouts:
            layouts = ['desktop', 'laptop']

        return sorted(layouts)
