#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _

import os
import importlib.util
import shutil

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), '..', 'config_base.py')
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
ConfigCommand = _module.ConfigCommand


class command(ConfigCommand):
    def __init__(self):
        super().__init__()

    def getDescription(self):
        return "Reset configuration to default settings"

    def run(self):
        self.presentText("WARNING: This will reset all settings to defaults")
        self.presentText("Creating backup before reset...")

        # Create backup first
        backupSuccess, backupMessage = self.backupConfig(announce=False)
        if not backupSuccess:
            self.presentText(
                "Failed to create backup. Reset cancelled for safety.",
                interrupt=False)
            self.playSound('Error')
            return

        # Find default configuration file
        defaultConfigPath = self.findDefaultConfig()

        if defaultConfigPath and os.path.isfile(defaultConfigPath):
            try:
                # Copy default configuration over current
                shutil.copy2(defaultConfigPath, self.settingsFile)

                # Reload the configuration
                self.reloadConfig()

                self.presentText(
                    "Configuration reset to defaults successfully",
                    interrupt=False,
                    flush=False)
                self.presentText(
                    "Please restart Fenrir for all changes to take effect",
                    interrupt=False,
                    flush=False)

            except Exception as e:
                self.presentText(
                    f"Failed to reset configuration: {
                        str(e)}", interrupt=False, flush=False)
        else:
            # Manually create basic default configuration
            self.createBasicDefaults()

    def findDefaultConfig(self):
        """Find the default configuration file"""
        possiblePaths = [
            '/usr/share/fenrirscreenreader/config/settings/settings.conf',
            '/etc/fenrirscreenreader/settings/settings.conf.default',
            os.path.join(
                os.path.dirname(
                    self.settingsFile),
                'settings.conf.default'),
            # Development path
            os.path.join(
                os.path.dirname(__file__),
                '..',
                '..',
                '..',
                '..',
                '..',
                '..',
                'config',
                'settings',
                'settings.conf')
        ]

        for path in possiblePaths:
            if os.path.isfile(path):
                return path

        return None

    def createBasicDefaults(self):
        """Create basic default configuration manually"""
        try:
            self.config.clear()

            # Basic speech defaults
            self.config['speech'] = {
                'enabled': 'True',
                'driver': 'speechdDriver',
                'rate': '0.5',
                'pitch': '0.5',
                'volume': '1.0',
                'autoReadIncoming': 'True'
            }

            # Basic sound defaults
            self.config['sound'] = {
                'enabled': 'True',
                'driver': 'genericDriver',
                'theme': 'default',
                'volume': '0.7'
            }

            # Basic keyboard defaults
            self.config['keyboard'] = {
                'driver': 'evdevDriver',
                'device': 'ALL',
                'keyboardLayout': 'desktop',
                'charEchoMode': '1',
                'wordEcho': 'False',
                'charDeleteEcho': 'True'
            }

            # Basic screen defaults
            self.config['screen'] = {
                'driver': 'vcsaDriver',
                'encoding': 'auto'
            }

            # Basic general defaults
            self.config['general'] = {
                'punctuationLevel': 'some',
                'debugLevel': '0',
                'numberOfClipboards': '50'
            }

            # Write the configuration
            with open(self.settingsFile, 'w') as configfile:
                self.config.write(configfile)

            self.presentText(
                "Basic default configuration created",
                interrupt=False,
                flush=False)
            self.presentText(
                "Please restart Fenrir for changes to take effect",
                interrupt=False,
                flush=False)

        except Exception as e:
            self.presentText(
                f"Failed to create default configuration: {
                    str(e)}", interrupt=False, flush=False)
