#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _

import os
import importlib.util

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), '..', 'config_base.py')
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
ConfigCommand = _module.ConfigCommand


class command(ConfigCommand):
    def __init__(self):
        super().__init__()

    def getDescription(self):
        return "Select sound theme"

    def run(self):
        currentTheme = self.getSetting('sound', 'theme', 'default')

        # Present current theme
        self.presentText(f"Current sound theme: {currentTheme}")

        # Look for available sound themes
        soundPaths = [
            '/usr/share/sounds',
            '/usr/share/fenrirscreenreader/sounds',
            os.path.expanduser('~/.local/share/fenrirscreenreader/sounds')
        ]

        availableThemes = self.getAvailableThemes(soundPaths)

        if len(availableThemes) > 1:
            # For this implementation, cycle through available themes
            try:
                currentIndex = availableThemes.index(currentTheme)
                nextIndex = (currentIndex + 1) % len(availableThemes)
                newTheme = availableThemes[nextIndex]
            except ValueError:
                # Current theme not found in available themes, use first one
                newTheme = availableThemes[0]

            success = self.setSetting('sound', 'theme', newTheme)

            if success:
                self.presentText(f"Sound theme changed to: {newTheme}")
                self.playSound('Accept')
            else:
                self.presentText("Failed to change sound theme")
                self.playSound('Error')
        else:
            self.presentText("Only default sound theme is available")
            self.playSound('Cancel')

    def getAvailableThemes(self, searchPaths):
        """Find available sound themes in the search paths"""
        themes = ['default']  # Always include default

        for path in searchPaths:
            if os.path.isdir(path):
                try:
                    for item in os.listdir(path):
                        fullPath = os.path.join(path, item)
                        if os.path.isdir(
                                fullPath) and item != 'default' and item not in themes:
                            # Check if it looks like a sound theme (contains
                            # sound files)
                            if self.isValidSoundTheme(fullPath):
                                themes.append(item)
                except Exception:
                    continue

        return themes

    def isValidSoundTheme(self, themePath):
        """Check if a directory contains sound files"""
        soundExtensions = ['.wav', '.ogg', '.mp3', '.flac']
        try:
            for file in os.listdir(themePath):
                if any(file.lower().endswith(ext) for ext in soundExtensions):
                    return True
        except Exception:
            pass
        return False
